/*
 * usage: pschecker [-1cdehiqsv] [-E excludefilename] [-o outfile]
 *                  [-p probehost] [-R time ] [-t timer] conffilename
 *   -1      run once then quit
 *   -c      display config and exit
 *   -d      add more debugging information (multiple -d are allowed)
 *   -e      detect running processes that are not in configuration file,
 *           and not in excluded list (if -E is used)
 *   -h      html output
 *   -i      output to be piped to itrs software
 *   -o      send output here instead of to stdout
 *   -p      host to send probe to
 *   -q      quiet, do not display if level=0
 *   -R      record until time
 *   -s      use syslog
 *   -t      rescan every timer second (default 30), if -1 not set
 *   -v      print version and exit
 * conffilename the name of the config file
 *
 * The safe way to terminate the program is SIGINT (ctrl-C) or SIGTERM
 * This utility was written by Philippe Charnier <charnier@xp11.frmug.org>
 */

/* the date is accurate but the revision number is not */
static const char rcsid[] =
  "$Id: pschecker.c,v 1.115 2012/03/08 08:28:40 phcha Exp $";

#include <arpa/inet.h>
#if defined(__FreeBSD__)
#include <sys/param.h>
#endif
#if defined(__linux__)
#include <asm/param.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netdb.h>
#include <netinet/in.h>
#include <ctype.h>
#include <libgen.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#if defined(__linux__)
#include <time.h>
#endif
#include <regex.h>
#include <unistd.h>

#define MAXBUFFERSIZE	4096*3	/* size of ps(1) output */
#define MAXLASTPROC	1024	/* # of processes */
#define MAXCONFIGLIST	1024	/* # of rules */
#define MAXEXCLUDELIST	256	/* # of excluded processes */
#define MAXPROBELIST	4

#define STATEUNKNOWN	0
#define STATERUNNING	1
#define STATEWRONGTIME	2
#define STATEEXCLUDED	3

#define FLAGUNKNOWN	0
#define FLAGHIGHPCPU	1
#define FLAGHIGHETIME	2
#define FLAGHIGHVSZ	4

int loop;
int lastproc;
int lastconfig;
int lastexclude;
int lastprobehost;
int time_hour_min;
int time_hour_min_sec;
int weekday;
char *category;
volatile int sig_terminate;
int timer;
char *conffilename;
char *excludefilename;
char hostname[MAXHOSTNAMELEN];
int record;
int processes, processesallowed, excludedprocesscount;
unsigned int virtualsize;

FILE *outfile;

struct sprobehost {
  char *probehost;
  int   sockfd;
  struct sockaddr_in name;
  int   public; /* -1 for global probehost (#probehost ... or -p probehost) */
};

struct sflags {
  int debug;
  int excluded;
  int html;
  int itrs;
  int once;
  int syslog;
  int displayconfig;
  int quiet;
  int sort_by_start_time;
};

struct ptag {
  char *pid;
  char *cmd;
  char *args;
  int   start_hour_min;
  double pcpu;
  unsigned int etime;
  unsigned int vsz; /* The total size of the process in virtual memory */
  int   loop;
  int   match;
  int   state;
};

struct ctag {
  char *cmd;
  char *args;
  int   start_hour_min;
  int   end_hour_min;
  int   alarm_level;
  int   weekday;
  double pcpu;
  unsigned int etime;
  unsigned int vsz;
  int   flags;
  int   match;
  int   state;
  char *comment;
  char *category;
  char *dirprefix;
  char *progname;
  char *mailto;
  int   delay;
};

struct etag {
  char *cmd;
  int initialcount;
  int remainingcount;
};

struct sflags f;
struct sprobehost probehostlist[MAXPROBELIST];
struct ptag proclist[MAXLASTPROC];
struct ctag configlist[MAXCONFIGLIST];
struct etag excludelist[MAXEXCLUDELIST];

/*
 * display error message and exit
 */
void
myerrx(int exitcode, char *msg)
{
  fprintf(stderr, "pschecker: %s\n", msg);
  exit(exitcode);
}

/*
 * remove trailing spaces and \n if any
 */
void
chomp(char *buffer)
{ char *ptr, *last_not_space;

  ptr = last_not_space = buffer;
  while (*ptr) {
    if (!isspace((int)*ptr))
      last_not_space = ptr;
    ptr++;
  }
  if (*last_not_space && !isspace((int)*last_not_space))
    last_not_space++;
  *last_not_space = '\0';
}

/*
 * send an UDP datagram to distant host
 */
void
sendprobemessage(char *msg, int public)
{
  int nmsg, i;

  for (i = 0; i < lastprobehost; i++) {
    if (public == -2 ||
        probehostlist[i].public == public ||
        probehostlist[i].public == -1) {
      nmsg = strlen(msg);
      if (sendto(probehostlist[i].sockfd, msg, nmsg, 0,
          (struct sockaddr *)&(probehostlist[i].name),
          sizeof(probehostlist[i].name)) != nmsg)
        myerrx(1, "sendto");
    }
  }
}

/*
 * locate the first non space in the string
 */
void
skipwhitespace(char **ptr)
{
  while(*ptr && isspace((int)**ptr)) (*ptr)++;
}

int
incomment(char *ptr)
{ static char *start = NULL;
  static char *stop  = NULL;
  int state;

  if (ptr == NULL) {
    start = stop = NULL;
    return (0);
  }

  if (start == NULL) {
    start = ptr;
    while (*start && strncmp("comment=", start, sizeof("comment=") - 1) != 0)
      start++;
    stop = start;
    if (*start) {
      start += sizeof("comment=") - 1;
      stop  = start + 1;
      state = 0;
      while (*stop && state != 2) {
        switch (state) {
          case 0: if (*stop == '\\') {
                    stop++;
                    state = 1; /* escape next char */
                  }
                  else if (*stop == *start) { state = 2; /* finish */}
                  else { stop++; }
                  break;
          case 1: stop++;
                  state = 0;
                  break;
        }
      }
    }
  }

  return (start < ptr && ptr < stop);
}

/*
 * get the next field in a space(s) separated fields of a string
 */
char *
shift (char **remain, int stopatspace)
{ char *result;

  while(**remain && isspace((int)**remain)) (*remain)++;
  result = *remain;
  if (4 <= f.debug)
    fprintf(outfile, "shift(): ##%s##\n", result);
  if (stopatspace) {
    incomment(NULL); /* initialise */
    while(**remain && !(!incomment(*remain) &&
          isspace((int)**remain)))
      (*remain)++;
    if (!incomment(*remain) && isspace((int)**remain)) {
      **remain = '\0';
      (*remain)++;
    }
  }
  else {
    while(**remain && **remain != '\n') (*remain)++;
    **remain = '\0';
  }
  return (result);
}

/*
 * search for a matching regex
 */
int
matchingextended(char *rgx, char *str)
{ int rv;
  regex_t reg;

  rv = regcomp(&reg, rgx, REG_EXTENDED|REG_NOSUB);
  if (rv != 0) {
    if (4 <= f.debug)
      fprintf(outfile, "matchingextended()/regcomp(): ##%s## and ##%s## returned %d\n", rgx, str, rv);
    myerrx(1, "regcomp");
  }
  rv = regexec(&reg, str, 0, NULL, 0);
  regfree(&reg);
  if (4 <= f.debug)
    fprintf(outfile, "matchingextended(): regex=##%s## and string=##%s## %s\n", rgx, str, (rv == 0) ? "[match]" : "[no match]");
  return (rv == 0);
}

/*
 * search for a matching regex
 */
int
isexcludedextended(char *cmd)
{ int i;

  for (i = 0; i < lastexclude; i++)
    if (0 < excludelist[i].remainingcount && matchingextended(excludelist[i].cmd, cmd)) {
      excludelist[i].remainingcount--;
      return (1);
    }
  return (0);
}

/*
 * ouput the contents
 */
void
dumpall(void)
{ int i;

  if (!(2 <= f.debug)) return;

  for (i = 0; i < lastproc; i++)
    fprintf(outfile, "  %04d 2400 %.1f %u %u %s %s %s\n",
            proclist[i].start_hour_min, proclist[i].pcpu,
            proclist[i].etime, proclist[i].vsz, proclist[i].pid, proclist[i].cmd, proclist[i].args);
}

/*
 * output an itrs software compatible report
 */
void
dumpconfigtoitrs(void)
{ int i, j;
  char *levelstr, *statusstr;

  levelstr = "(null)";
  statusstr = "(null)";
  printf ("index,comment,level,status,category,pid,command,args,start time,end time..\n");
  for (i = 0; i < lastconfig; i++) {
    if (configlist[i].start_hour_min == configlist[i].end_hour_min)
      continue;
    if ((configlist[i].weekday & weekday) == 0)
      continue;
    switch(configlist[i].state) {
      case STATEUNKNOWN : /* still unknown, may be missing */
                          if (time_hour_min < configlist[i].start_hour_min
                              || configlist[i].end_hour_min <= time_hour_min) {
                            levelstr = "OK";
                            statusstr = "not checked";
                            break;
                          }

                          if (50 < configlist[i].alarm_level)
                            levelstr = "CRITICAL";
                          else if (20 < configlist[i].alarm_level)
                            levelstr = "WARNING";
                          else
                            levelstr = "STANDBY";
                          statusstr = "missing";
                          if (configlist[i].alarm_level == 0) {
                            levelstr = "OK";
                            statusstr = "not checked";
                          }
                          break;
      case STATERUNNING :
                          levelstr = "OK";
                          statusstr = "running";
                          if (configlist[i].flags & FLAGHIGHETIME) {
                            levelstr = "WARNING";
                            statusstr = "high etime";
                          }
                          if (configlist[i].flags & FLAGHIGHVSZ) {
                            levelstr = "WARNING";
                            statusstr = "high vsz";
                          }
                          if (configlist[i].flags & FLAGHIGHPCPU) {
                            levelstr = "WARNING";
                            statusstr = "high pcpu";
                          }
                          break;
      case STATEWRONGTIME :
                          levelstr = "WARNING";
                          statusstr = "wrong time";
                          break;
    }
    if (configlist[i].alarm_level != 0 || f.quiet == 0)
      printf ("%d,%s,%s,%s,%c-%s,%s,%s,%s,%04d,%04d\n",
            i, configlist[i].comment, levelstr, statusstr,
            levelstr[0], configlist[i].category, (configlist[i].match != -1) ? proclist[configlist[i].match].pid : "n/a", configlist[i].cmd,
            strchr(configlist[i].args, ',') ? "(not displayed)" : configlist[i].args,
            configlist[i].start_hour_min, configlist[i].end_hour_min);
  }
  /* Now, display unknown running processes. Do not display args which can
   * contain commas.
   */
  if (f.excluded) {
    for (j = 0; j < lastproc; j++) {
      if (proclist[j].state == STATEUNKNOWN) {
        printf ("%d,%s,%s,%s,%s,%s,%s,%s,%04d,%04d\n", i, "n/a", "WARNING",
                "unknown", "W-n/a", proclist[j].pid, proclist[j].cmd,
                strchr(proclist[j].args, ',')
                ? "(not displayed)" : proclist[j].args, 0, 0);
        i++;
      }
    }
  }
}

/*
 * html output
 */
void
dumpconfigtohtml(void)
{ int i, j;
  char *levelcolor, *levelstr, *statusstr;

  printf("<CENTER>\n"
         "<TABLE BORDER=1 CELLSPACING=0 CELLPADDING=5>\n"
         "<TR>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">index</FONT></TD>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">comment</FONT></TD>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">level</FONT></TD>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">status</FONT></TD>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">category</FONT></TD>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">pid</FONT></TD>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">command</FONT></TD>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">args</FONT></TD>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">start time</FONT></TD>\n"
         "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\"><FONT COLOR=\"#990099\">end time</FONT></TD>\n"
         "</TR>\n");

  for (i = 0; i < lastconfig; i++) {
    if (configlist[i].start_hour_min == configlist[i].end_hour_min)
      continue;
    if ((configlist[i].weekday & weekday) == 0)
      continue;

    levelcolor = "#990099";
    levelstr = "(null)";
    statusstr = "(null)";

    switch(configlist[i].state) {
      case STATEUNKNOWN : /* still unknown, may be missing */
                          if (time_hour_min < configlist[i].start_hour_min
                              || configlist[i].end_hour_min <= time_hour_min) {
                            levelcolor = "#00FF00";
                            levelstr = "OK";
                            statusstr = "not checked";
                            break;
                          }

                          if (50 < configlist[i].alarm_level) {
                            levelcolor = "#FF0000";
                            levelstr = "CRITICAL";
                          }
                          else if (20 < configlist[i].alarm_level) {
                            levelcolor = "#FFFF00";
                            levelstr = "WARNING";
                          }
                          else {
                            levelcolor = "#00FF00";
                            levelstr = "STANDBY";
                          }
                          statusstr = "missing";
                          if (configlist[i].alarm_level == 0) {
                            levelcolor = "#00FF00";
                            levelstr = "OK";
                            statusstr = "not checked";
                          }
                          break;
      case STATERUNNING :
                          levelcolor = "#00FF00";
                          levelstr = "OK";
                          statusstr = "running";
                          if (configlist[i].flags & FLAGHIGHETIME) {
                            levelcolor = "#FFFF00";
                            levelstr = "WARNING";
                            statusstr = "high etime";
                          }
                          if (configlist[i].flags & FLAGHIGHVSZ) {
                            levelcolor = "#FFFF00";
                            levelstr = "WARNING";
                            statusstr = "high vsz";
                          }
                          if (configlist[i].flags & FLAGHIGHPCPU) {
                            levelcolor = "#FFFF00";
                            levelstr = "WARNING";
                            statusstr = "high pcpu";
                          }
                          break;
      case STATEWRONGTIME :
                          levelcolor = "#FFFF00";
                          levelstr = "WARNING";
                          statusstr = "wrong time";
                          break;
    }
    if (configlist[i].alarm_level != 0 || f.quiet == 0)
      printf("<TR>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%d</TD>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"%s\">%s</TD>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%04d</TD>\n"
           "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%04d</TD>\n"
           "</TR>\n",
            i, configlist[i].comment, levelcolor, levelstr, statusstr,
            configlist[i].category, (configlist[i].match != -1) ? proclist[configlist[i].match].pid : "n/a", configlist[i].cmd,
            (strlen(configlist[i].args) != 0) ? configlist[i].args : "&nbsp;",
            configlist[i].start_hour_min, configlist[i].end_hour_min);
  }
  /* Now, display unknown running processes. Do not display args which can
   * contain commas.
   */
  if (f.excluded) {
    for (j = 0; j < lastproc; j++) {
      if (proclist[j].state == STATEUNKNOWN) {
        printf("<TR>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%d</TD>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#0000FF\">%s</TD>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%s</TD>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%04d</TD>\n"
               "<TD ALIGN=LEFT BGCOLOR=\"#7fffd4\">%04d</TD>\n"
               "</TR>\n",
                i, "n/a", "WARNING", "unknown", "n/a", proclist[j].pid, proclist[j].cmd,
                (strlen(proclist[j].args) != 0) ? proclist[j].args : "&nbsp;",
                0, 0);
        i++;
      }
    }
  }
  printf("</TABLE></CENTER>\n");
}

/*
 * convert a binary string "0111100" representing a mask of the days in a
 * week to an integer
 */
int
atobin(char *str)
{ int result = 0;
  char *ptr;

  if (strlen(str) != 7)
    myerrx(1, "incorrect weekday length");

  ptr = str;
  while(*ptr && (*ptr == '1' || *ptr == '0'))
    ptr++;

  if (*ptr)
    myerrx(1, "incorrect weekday definition");

  ptr = str;
  while (*ptr) {
    result *= 2;
    result += (*ptr == '1') ? 1 : 0;
    ptr++;
  }
  return (result);
}

/*
 * parse config file for directives, comments, blank lines
 */
void readexcludeentry(char *);

int
commentorblank(char *buffer)
{ char *from, *to;
  int state;

#define READDIRECT(x) if (strncmp(buffer, x, sizeof(x) - 1) == 0)

  chomp(buffer);

  /* read directives, override command line options */
  READDIRECT("#excluded_flag 0") {
    f.excluded = 0;
  }
  READDIRECT("#excluded_flag off") {
    f.excluded = 0;
  }
  READDIRECT("#excluded_flag 1") {
    f.excluded = 1;
  }
  READDIRECT("#excluded_flag on") {
    f.excluded = 1;
  }

  READDIRECT("#quiet_flag 0") {
    f.quiet = 0;
  }
  READDIRECT("#quiet_flag off") {
    f.quiet = 0;
  }
  READDIRECT("#quiet_flag 1") {
    f.quiet = 1;
  }
  READDIRECT("#quiet_flag on") {
    f.quiet = 1;
  }

  READDIRECT("#sort_by_start_time 0") {
    f.sort_by_start_time = 0;
  }
  READDIRECT("#sort_by_start_time off") {
    f.sort_by_start_time = 0;
  }
  READDIRECT("#sort_by_start_time 1") {
    f.sort_by_start_time = 1;
  }
  READDIRECT("#sort_by_start_time on") {
    f.sort_by_start_time = 1;
  }

  READDIRECT("#excluded_rule") {
    readexcludeentry(&buffer[sizeof("#excluded_rule")]);
  }

  READDIRECT("#probehost") {
    /*
     * e.g. #probehost host_where_servchecker_is_listening
     * trivial syntax required: one space before hostname, no space after
     */
    if (lastprobehost == MAXPROBELIST)
      myerrx(1, "increase MAXPROBELIST");
    probehostlist[lastprobehost].public = -1;
    probehostlist[lastprobehost].probehost =
                                 strdup(&buffer[sizeof("#probehost")]);
    lastprobehost++;
  }

  READDIRECT("#weekday") {
    weekday = atobin(&buffer[sizeof("#weekday")]);
  }

  READDIRECT("#vsz") {
    virtualsize = atoi(&buffer[sizeof("#vsz")]);
  }

  READDIRECT("#processes") {
    processesallowed = atoi(&buffer[sizeof("#processes")]);
  }

  READDIRECT("#excludedprocesscount") {
    excludedprocesscount = atoi(&buffer[sizeof("#excludedprocesscount")]);
  }

  READDIRECT("#category") {
    free(category);
    category = strdup(&buffer[sizeof("#category")]);
  }

#undef READDIRECT

  /*
   * if # is escaped, replace \# with # and continue. If # is not escaped
   * thus starting a comment, remove after it.
   */
  if (4 <= f.debug)
    fprintf(outfile, "commentorblank(): ##%s##", buffer);

  from = to = buffer;
  state = 0;
  while (*from) {
    switch (state) {
    case 0 : if (*from == '#') { *from = '\0'; /* finish */ }
             else if (*from == '\\') {
               from++;
               state = 1; /* escape next char */
             }
             else { *to = *from; to++; from++; }
             break;
    case 1 : if (*from != '#')
               *to++ = '\\';
             *to = *from; to++; from++;
             state = 0;
             break;
    }
  }
  *to = '\0';

  if (4 <= f.debug)
    fprintf(outfile, " replaced with ##%s##\n", buffer);

  return (*buffer == '\0');
}

/*
 * read line for excluded processes
 */
void
readexcludeentry(char *buffer)
{ char *remainder;

  if (commentorblank(buffer))
    return;
  if (lastexclude == MAXEXCLUDELIST)
    myerrx(1, "increase MAXEXCLUDELIST");
  remainder = buffer;
  skipwhitespace(&remainder);
  excludelist[lastexclude].cmd =
      strdup(shift(&remainder, 0 /* do not stop at space */));
  excludelist[lastexclude].initialcount = excludedprocesscount;
  lastexclude++;
}

/*
 * read file for excluded processes
 */
void
readexcludefile(void)
{ char buffer[1024];
  FILE *stream;

  if (excludefilename == NULL)
    return;

  stream = fopen(excludefilename, "r");
  if (stream == NULL) {
    fprintf(stderr, "pschecker: can not read file %s\n", excludefilename);
    exit(1);
  }
  while ((fgets(buffer, sizeof(buffer) - 1, stream)) != NULL)
    readexcludeentry(buffer);
  fclose(stream);
}

/*
 * search where substring "..." is located in the strings
 */
int
find3dots(char *a, char *b)
{

  /* shift until the first diff */
  while (*a != '\0' && *a == *b) {
   a++;
   b++;
  }

  if (*b == '\0') return (1);
  if (*a == '\0') return (-1);

  if (strncmp("...", b, sizeof("...")) == 0) return (1);
  if (strncmp("...", a, sizeof("...")) == 0) return (-1);

  return (0);
}

char *
strdupcomment(char *ptr)
{ char *result;
  char *from, *to;
  int state;

  result = (char *)malloc(strlen(ptr) + 1);
  if (result == NULL)
    return (NULL);

  if (4 <= f.debug)
    fprintf(outfile, "strdupcomment(): ##%s##\n", ptr);

  from = ptr;
  to = result;

  from  = ptr + 1;
  state = 0;
  while (*from && state != 2) {
    switch (state) {
    case 0: if (*from == '\\') {
              from++;
              state = 1; /* escape next char */
            }
            else if (*from == *ptr) { state = 2; /* finish */}
            else { *to = *from; to++; from++; }
            break;
    case 1: *to = *from; to++; from++;
            state = 0;
            break;
    }
  }
  *to = '\0';
  return (result);
}

/*
 * display config file
 */
void
displayconfig(void)
{ int i;

  for (i = 0; i < lastconfig; i++) {
    /* need -cc to display if alarm_level is 0 */
    if (1 < f.displayconfig || configlist[i].alarm_level != 0)
      printf("%s %d %d %d %s %s %s\n",
        configlist[i].cmd, configlist[i].start_hour_min,
        configlist[i].end_hour_min, configlist[i].alarm_level,
        configlist[i].dirprefix, configlist[i].category, configlist[i].progname);
  }
}

/*
 * get memory size in the same format as in ps -o vsz (in KB)
 */
unsigned int
getmemorysize(void)
{ unsigned int i;

  i = (sysconf(_SC_PAGESIZE) >> 10) * sysconf(_SC_PHYS_PAGES);
  if (1 <= f.debug)
    fprintf(outfile, "pschecker: memory size found in KB: %u\n", i);
  return (i);
}

/*
 * read config file
 */
void
readconffile(void)
{ char *remainder, *ptr, *ptr2, *result;
  char buffer[1024];
  int found, i, j;
  FILE *stream;
  int repeat;

  lastconfig = 0;

  /*
   * cmd and args are obtained from:
   * Solaris /usr/bin/ps -eo args
   * FreeBSD /bin/ps -axo command
   * Linux   /bin/ps -eo command
   */
  stream = fopen(conffilename, "r");
  if (stream == NULL) {
    fprintf(stderr, "pschecker: can not read file %s\n", conffilename);
    exit(1);
  }
  while ((fgets(buffer, sizeof(buffer) - 1, stream)) != NULL) {
    if (commentorblank(buffer))
      continue;
    if (lastconfig == MAXCONFIGLIST)
      myerrx(1, "increase MAXCONFIGLIST");
    remainder = buffer;
    skipwhitespace(&remainder);
    configlist[lastconfig].start_hour_min = atoi(shift(&remainder, 1 /* stop at first space */));
    configlist[lastconfig].end_hour_min = atoi(shift(&remainder, 1 /* stop at first space */));

    configlist[lastconfig].alarm_level = 90; /* default value */
    configlist[lastconfig].weekday = weekday; /* default value */
    configlist[lastconfig].comment = strdup("n/a");
    configlist[lastconfig].category = strdup(category); /* default value */
    configlist[lastconfig].dirprefix = strdup("/"); /* default value */
    configlist[lastconfig].progname = strdup("n/a"); /* default value */
    configlist[lastconfig].mailto = strdup("n/a"); /* default value */
    configlist[lastconfig].pcpu = 70.0; /* default value */
    configlist[lastconfig].etime = (configlist[lastconfig].start_hour_min != 0 || configlist[lastconfig].end_hour_min != 2400); /* default value */
    configlist[lastconfig].vsz = virtualsize; /* default value */
    configlist[lastconfig].delay = 0; /* default value */
    ptr = ptr2 = strdup(shift(&remainder, 1 /* stop at first space */));
    repeat = 1;
    while ((result = strtok(ptr2, ","))) {
      ptr2 = NULL;
      found = 0;
      if (found == 0 && strncmp("level=", result, sizeof("level=") - 1) == 0) {
        configlist[lastconfig].alarm_level = atoi(&result[sizeof("level=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("weekday=", result, sizeof("weekday=") - 1) == 0) {
        configlist[lastconfig].weekday = atobin(&result[sizeof("weekday=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("comment=", result, sizeof("comment=") - 1) == 0) {
        free(configlist[lastconfig].comment);
        configlist[lastconfig].comment = strdupcomment(&result[sizeof("comment=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("category=", result, sizeof("category=") - 1) == 0) {
        free(configlist[lastconfig].category);
        configlist[lastconfig].category = strdup(&result[sizeof("category=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("dirprefix=", result, sizeof("dirprefix=") - 1) == 0) {
        free(configlist[lastconfig].dirprefix);
        configlist[lastconfig].dirprefix = strdup(&result[sizeof("dirprefix=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("progname=", result, sizeof("progname=") - 1) == 0) {
        free(configlist[lastconfig].progname);
        configlist[lastconfig].progname = strdup(&result[sizeof("progname=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("mailto=", result, sizeof("mailto=") - 1) == 0) {
        free(configlist[lastconfig].mailto);
        configlist[lastconfig].mailto = strdup(&result[sizeof("mailto=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("probehost=", result, sizeof("probehost=") - 1) == 0) {
        if (lastprobehost == MAXPROBELIST)
          myerrx(1, "increase MAXPROBELIST");
        probehostlist[lastprobehost].public = lastconfig;
        probehostlist[lastprobehost].probehost =
            strdup(&result[sizeof("probehost=") - 1]);
        lastprobehost++;
        found = 1;
      }
      if (found == 0 && strncmp("pcpu=", result, sizeof("pcpu=") - 1) == 0) {
        configlist[lastconfig].pcpu = atof(&result[sizeof("pcpu=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("etime=", result, sizeof("etime=") - 1) == 0) {
        configlist[lastconfig].etime = atoi(&result[sizeof("etime=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("vsz=", result, sizeof("vsz=") - 1) == 0) {
        configlist[lastconfig].vsz = atoi(&result[sizeof("vsz=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("repeat=", result, sizeof("repeat=") - 1) == 0) {
        repeat = atoi(&result[sizeof("repeat=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("delay=", result, sizeof("delay=") - 1) == 0) {
        configlist[lastconfig].delay = atoi(&result[sizeof("delay=") - 1]);
        found = 1;
      }
      if (found == 0) {
        if (1 <= f.debug)
          fprintf(outfile, "pschecker: unknown option field: %s\n", result);
        if (1 <= f.syslog)
          syslog(LOG_ERR, "pschecker: unknown option field: %s", result);
        /* do not abort here, to make old config work exit(1); */
      }
    }
    free(ptr);

    configlist[lastconfig].cmd = strdup(shift(&remainder, 1 /* stop at first space */));
    configlist[lastconfig].args= strdup(shift(&remainder, 0 /* do not stop at space */));
    configlist[lastconfig].flags = FLAGUNKNOWN;
    configlist[lastconfig].match = -1;
    configlist[lastconfig].state = STATEUNKNOWN;
    lastconfig++;
    while (1 < repeat) {
      if (lastconfig == MAXCONFIGLIST)
        myerrx(1, "increase MAXCONFIGLIST");
      configlist[lastconfig].start_hour_min = configlist[lastconfig-1].start_hour_min;
      configlist[lastconfig].end_hour_min = configlist[lastconfig-1].end_hour_min;
      configlist[lastconfig].alarm_level = configlist[lastconfig-1].alarm_level;
      configlist[lastconfig].weekday = configlist[lastconfig-1].weekday;
      configlist[lastconfig].pcpu = configlist[lastconfig-1].pcpu;
      configlist[lastconfig].etime = configlist[lastconfig-1].etime;
      configlist[lastconfig].vsz = configlist[lastconfig-1].vsz;
      configlist[lastconfig].cmd = strdup(configlist[lastconfig-1].cmd);
      configlist[lastconfig].args = strdup(configlist[lastconfig-1].args);
      configlist[lastconfig].flags = configlist[lastconfig-1].flags;
      configlist[lastconfig].match = configlist[lastconfig-1].match;
      configlist[lastconfig].comment = strdup(configlist[lastconfig-1].comment);
      configlist[lastconfig].category = strdup(configlist[lastconfig-1].category);
      configlist[lastconfig].dirprefix = strdup(configlist[lastconfig-1].dirprefix);
      configlist[lastconfig].progname = strdup(configlist[lastconfig-1].progname);
      configlist[lastconfig].mailto = strdup(configlist[lastconfig-1].mailto);
      configlist[lastconfig].state = configlist[lastconfig-1].state;
      configlist[lastconfig].delay = configlist[lastconfig-1].delay;
      repeat--;
      lastconfig++;
    }
  }
  fclose(stream);

  /*
   * reduce matching capabilities in config file
   * example:       config list           process list
   *         (1)cmd=a pcpu=50 etime=60  (a)cmd=a pcpu=70 etime=50
   *         (2)cmd=a pcpu=90 etime=40
   * program can choose (1)-(a) and alarm because of pcpu
   * or choose (2)-(a) and alarm because of etime. For the time beeing, it
   * will choose the first match (1)-(a).
   *
   * example:
   *         (1)cmd=a pcpu=70           (a)cmd=a pcpu=75
   *         (2)cmd=a pcpu=80           (b)cmd=a pcpu=50
   * program will alarm on pcpu because (a) does not match (1). The correct
   * bijection is (a)-(2) and (b)-(1). For the time beeing, it will choose
   * the first match (1)-(a) even if (a)-(2) is a best match.
   *
   * We check that for a given cmd=a, all pcpu are the same, all etime are
   * the same.
   */
  for (i = 0; i < lastconfig; i++) {
    if ((configlist[i].weekday & weekday) == 0)
      continue;
    for (j = i + 1; j < lastconfig; j++) {
      if ((configlist[j].weekday & weekday) == 0)
        continue;
      if ((strcmp(configlist[i].cmd, configlist[j].cmd) == 0) &&
          find3dots(configlist[i].args, configlist[j].args) < 0) {
        if (1 <= f.debug)
          fprintf(outfile, "incorrect ordering: %s %s\n",
                  configlist[i].cmd, configlist[i].args);
        if (1 <= f.syslog)
          syslog(LOG_ERR, "incorrect ordering: %s %s",
                 configlist[i].cmd, configlist[i].args);
      }
      if (strcmp(configlist[i].cmd, configlist[j].cmd) == 0 &&
          strcmp(configlist[i].args, configlist[j].args) == 0) {
        if (configlist[i].pcpu != configlist[j].pcpu) {
          if (1 <= f.debug)
            fprintf(outfile, "incorrect pcpu=%.1f != %.1f found above\n",
                    configlist[j].pcpu, configlist[i].pcpu);
          if (1 <= f.syslog)
            syslog(LOG_ERR, "incorrect pcpu=%.1f != %.1f found above",
                   configlist[j].pcpu, configlist[i].pcpu);
        }
        if (configlist[i].etime != configlist[j].etime) {
          if (1 <= f.debug)
            fprintf(outfile, "incorrect etime=%d != %d found above\n",
                    configlist[j].etime, configlist[i].etime);
          if (1 <= f.syslog)
            syslog(LOG_ERR, "incorrect etime=%d != %d found above",
                   configlist[j].etime, configlist[i].etime);
        }
        if (configlist[i].vsz != configlist[j].vsz) {
          if (1 <= f.debug)
            fprintf(outfile, "incorrect vsz=%d != %d found above\n",
                    configlist[j].vsz, configlist[i].vsz);
          if (1 <= f.syslog)
            syslog(LOG_ERR, "incorrect vsz=%d != %d found above",
                   configlist[j].vsz, configlist[i].vsz);
        }
        if (configlist[i].alarm_level < configlist[j].alarm_level) {
          if (1 <= f.debug)
            fprintf(outfile, "incorrect alarm_level=%d -gt %d found above\n",
                    configlist[j].alarm_level, configlist[i].alarm_level);
          if (1 <= f.syslog)
            syslog(LOG_ERR, "incorrect alarm_level=%d -gt %d found above",
                   configlist[j].alarm_level, configlist[i].alarm_level);
        }
      }
    }
  }
}

/*
 * reset data structures
 */
void
resetstate(void)
{ int i;

  for (i = 0; i < lastproc; i++)
    proclist[i].state = STATEUNKNOWN;
  for (i = 0; i < lastconfig; i++) {
    configlist[i].flags = FLAGUNKNOWN;
    configlist[i].state = STATEUNKNOWN;
  }
  for (i = 0; i < lastexclude; i++) {
    excludelist[i].remainingcount = excludelist[i].initialcount;
  }
}

/*
 * find if the two strings match. The substring "..." match everything
 * until the end of the string. The substring "-*" remove options in
 * the command line.
 */
int
matchingargs(char *rgx, char *str)
{
  if (4 <= f.debug)
    fprintf(outfile, "matchingargs(): ##%s## and ##%s##\n", rgx, str);

  if (0 == strncmp("regex:", rgx, sizeof("regex:") - 1))
    return(matchingextended(&rgx[sizeof("regex:") - 1], str));

  while (*str == '-' && 0 == strncmp("-*", rgx, sizeof("-*") - 1)) {
    if (0 == strncmp("-classpath ", str, sizeof("-classpath ") - 1)) {
      /* classpath is followed by a space */
      str += sizeof("-classpath ") - 1;
      while (*str == ' ') str++;
    } 
    if (0 == strncmp("-cp ", str, sizeof("-cp ") - 1)) {
      /* classpath is followed by a space */
      str += sizeof("-cp ") - 1;
      while (*str == ' ') str++;
    } 
    while (*str != '\0' && *str != ' ') str++;
    while (*str == ' ') str++;

    if (4 <= f.debug)
      fprintf(outfile, "matchingargs(): ##%s## and ##%s##\n", rgx, str);
  }
  if (0 == strncmp("-*", rgx, sizeof("-*") - 1)) {
    rgx += sizeof("-*") - 1;
    while (*rgx == ' ') rgx++;
  }

  if (4 <= f.debug)
    fprintf(outfile, "matchingargs(): ##%s## and ##%s##\n", rgx, str);

  while (*rgx != '\0' && *str != '\0' && *rgx == *str &&
         strncmp("...", rgx, sizeof("..."))) {
    rgx++;
    str++;
  }
  if (4 <= f.debug)
    fprintf(outfile, "matchingargs(): ##%s## and ##%s##", rgx, str);
  if (*rgx == '\0' && *str == '\0') {
    if (4 <= f.debug)
      fprintf(outfile, " will match\n");
    return (1);
  }

  if (*rgx == '\0' || *str == '\0') {
    if (4 <= f.debug)
      fprintf(outfile, " will not match\n");
    return (0);
  }

  if (strncmp("...", rgx, sizeof("...")) == 0) {
    if (4 <= f.debug)
      fprintf(outfile, " will match\n");
    return (1);
  }

  if (4 <= f.debug)
    fprintf(outfile, " will not match\n");
  return (0);
}

/*
 * find if the two strings match.
 */
int
matchingcmd(char *rgx, char *str)
{
  if (4 <= f.debug)
    fprintf(outfile, "matchingcmd(): ##%s## and ##%s##\n", rgx, str);

  if (0 == strncmp("regex:", rgx, sizeof("regex:") - 1))
    return(matchingextended(&rgx[sizeof("regex:") - 1], str));

  return (0 == strcmp(rgx, str));
}

/*
 * create a matching relation between configlist[] and proclist[]
 */
int
checkmatching(int c, int i)
{ int ii;

  if ((configlist[c].weekday & weekday) == 0)
    return (0);

  if (3 <= f.debug) {
    for (ii = 0; ii < lastproc; ii++)
      if (proclist[ii].match != -1 &&
          configlist[proclist[ii].match].match != ii)
        fprintf(outfile, "checkmatching(): failed\n");
    for (ii = 0; ii < lastconfig; ii++)
      if (configlist[ii].match != -1 &&
          proclist[configlist[ii].match].match != ii)
        fprintf(outfile, "checkmatching(): failed\n");
  }

  if (configlist[c].match == i && proclist[i].match == c)
    return (1);

  if (configlist[c].match == -1 && proclist[i].match == -1 &&
      matchingcmd(configlist[c].cmd, proclist[i].cmd) &&
      matchingargs(configlist[c].args, proclist[i].args)) {
    configlist[c].match = i;
    proclist[i].match = c;
    return (1);
  }
  return (0);
}

/*
 * find a match
 */
int
isrunning(int c, int newstate)
{ int i;

  for (i = 0; i < lastproc; i++) {
    if (proclist[i].state == STATEUNKNOWN && checkmatching(c, i)) {
      proclist[i].state = newstate;
      configlist[c].state = newstate;
      return (1);
    }
  }
  return (0);
}

/*
 * tag high pcpu process
 */
int
ishighpcpu(int c)
{ int i;

  for (i = 0; i < lastproc; i++) {
    if (proclist[i].state == STATERUNNING &&
        configlist[c].pcpu <= proclist[i].pcpu &&
        checkmatching(c, i)) {
      configlist[c].flags |= FLAGHIGHPCPU;
      return (1);
    }
  }
  return (0);
}

/*
 * tag high etime process
 */
int
ishighetime(int c)
{ int i;

  if (configlist[c].etime == 0)
    return (0);

  for (i = 0; i < lastproc; i++) {
    if (proclist[i].state != STATERUNNING)
      continue;
    if (configlist[c].etime == 1) {
      if (time_hour_min_sec <= proclist[i].etime && checkmatching(c, i)) {
        configlist[c].flags |= FLAGHIGHETIME;
        return (1);
      }
    }
    else {
      if (configlist[c].etime <= proclist[i].etime && checkmatching(c, i)) {
        configlist[c].flags |= FLAGHIGHETIME;
        return (1);
      }
    }
  }
  return (0);
}

/*
 * tag high vsz process
 */
int
ishighvsz(int c)
{ int i;

  if (configlist[c].vsz == 0)
    return (0);

  for (i = 0; i < lastproc; i++) {
    if (proclist[i].state != STATERUNNING)
      continue;
    if (configlist[c].vsz <= proclist[i].vsz && checkmatching(c, i)) {
      configlist[c].flags |= FLAGHIGHVSZ;
      return (1);
    }
  }
  return (0);
}

/*
 * hide excluded processes unless in config file. Example:
 *   /usr/sbin/progname is in the config file
 *   ^/usr/sbin/.*$ is an excluded pattern
 *   if another /usr/sbin/progname is running, it must be shown as UNKNOWN
 */
void
markexcluded(void)
{ int i, c;

  for (i = 0; i < lastproc; i++)
    if (proclist[i].state == STATEUNKNOWN &&
        isexcludedextended(proclist[i].cmd)) {
      for (c = 0; c < lastconfig; c++) {
        if ((configlist[c].weekday & weekday) == 0)
          continue;
        if (matchingcmd(configlist[c].cmd, proclist[i].cmd) &&
            matchingargs(configlist[c].args, proclist[i].args))
          break;
      }
      if (c == lastconfig)
        proclist[i].state = STATEEXCLUDED;
    }
}

/*
 * show left processes
 */
void
unknown(void)
{ int i;
  char msg[1024];

  for (i = 0; i < lastproc; i++) {
    if (proclist[i].state == STATEUNKNOWN) {
      if (1 <= f.debug)
        fprintf(outfile, "unknown running process: %s %s\n", proclist[i].pid, proclist[i].cmd);
      else {
        sprintf(msg, ":h=%s:a=P-UNKNOWN_RUNNING:l=%d:pid=%s:x=%s:", hostname,
                                     timer, proclist[i].pid, proclist[i].cmd);
        sendprobemessage(msg, -1);
      }
      if (1 <= f.syslog)
        syslog(LOG_ERR, "unknown running process: %s %s", proclist[i].pid, proclist[i].cmd);
    }
  }
}

int
addtime(int c)
{ int minutes, hours;

  minutes = (configlist[c].start_hour_min % 100) + configlist[c].delay;
  hours   = configlist[c].start_hour_min / 100;

  while (60 <= minutes) {
    hours++;
    minutes-=60;
  }

  return ((24 <= hours) ?  2400 : hours * 100 + minutes);
}

/*
 * show missing processes
 */
void
missing(void)
{ int i;
  char msg[1024];

  for (i = 0; i < lastconfig; i++) {
    if (configlist[i].start_hour_min == configlist[i].end_hour_min)
      continue;
    if ((configlist[i].weekday & weekday) == 0)
      continue;

    if (addtime(i) <= time_hour_min &&
        time_hour_min < configlist[i].end_hour_min) {
      /* process should be running */
      if (!isrunning(i, STATERUNNING)) {
        if (1 <= f.debug) {
          if (configlist[i].alarm_level != 0)
            fprintf(outfile, "missing at %06d: %s %s {%04d %04d %d}\n",
                    time_hour_min_sec, configlist[i].cmd, configlist[i].args,
                    configlist[i].start_hour_min, configlist[i].end_hour_min,
                    configlist[i].alarm_level);
        }
        else {
          /*
           * use time_hour_min_sec to be sure that all messages will be
           * differents. They will not be buffered by a distant syslogd
           */
          if (configlist[i].alarm_level != 0) {
            if (1 <= f.syslog)
              syslog(LOG_ERR, "missing at %06d: %s %s {%04d %04d %d}",
                     time_hour_min_sec, configlist[i].cmd, configlist[i].args,
                     configlist[i].start_hour_min, configlist[i].end_hour_min,
                     configlist[i].alarm_level);
            if (90 < configlist[i].alarm_level)
              sprintf(msg, ":h=%s:a=P-MISSING-URGENT:l=%d:c=%s:m=%s:x=%s:", hostname,
                    timer, configlist[i].category, configlist[i].mailto, configlist[i].cmd);
            else if (20 < configlist[i].alarm_level)
              sprintf(msg, ":h=%s:a=P-MISSING:l=%d:c=%s:x=%s:", hostname,
                    timer, configlist[i].category, configlist[i].cmd);
            else
              sprintf(msg, ":h=%s:a=P-MISSING-STANDBY:l=%d:c=%s:x=%s:", hostname,
                    timer, configlist[i].category, configlist[i].cmd);
            sendprobemessage(msg, i);
          }
          if (matchingcmd(configlist[i].cmd, "/usr/sbin/cron")) {
            if (1 <= f.syslog)
              syslog(LOG_ERR, "restarting /usr/sbin/cron");
            system("/files/misc/bin/sudo -u root /bin/sh /etc/init.d/cron start");
          }
#ifdef SOLARIS551
          if (matchingcmd(configlist[i].cmd, "/usr/sbin/in.routed")) {
            if (1 <= f.syslog)
              syslog(LOG_ERR, "restarting /usr/sbin/in.routed");
            system("/files/misc/bin/sudo -u root /usr/sbin/in.routed -q");
          }
#endif
          if (matchingcmd(configlist[i].cmd, "/usr/sbin/syslogd")) {
            if (1 <= f.syslog)
              syslog(LOG_ERR, "restarting /usr/sbin/syslogd");
            system("/files/misc/bin/sudo -u root /bin/sh /etc/init.d/syslog start");
          }
        }
      }
    }
  }
}

/*
 * show running processes during wrong time
 */
void
running(void)
{ int i;
  char msg[1024];

  for (i = 0; i < lastconfig; i++) {
    if (configlist[i].start_hour_min == configlist[i].end_hour_min)
      continue;
    if ((configlist[i].weekday & weekday) == 0)
      continue;

    if (time_hour_min < configlist[i].start_hour_min
       || configlist[i].end_hour_min <= time_hour_min) {
      /* process should not be running */
      if (isrunning(i, STATEWRONGTIME)) {
        if (1 <= f.debug) {
          if (configlist[i].alarm_level != 0)
            fprintf(outfile, "running outside of [%04d,%04d] at %06d: %s %s\n",
                    configlist[i].start_hour_min, configlist[i].end_hour_min,
                    time_hour_min_sec, configlist[i].cmd, configlist[i].args);
        }
        else {
          /*
           * use time_hour_min_sec to be sure that all messages will be
           * differents. They will not be buffered by a distant syslogd
           */
          if (configlist[i].alarm_level != 0) {
            if (1 <= f.syslog)
              syslog(LOG_ERR, "running outside of [%04d,%04d] at %06d: %s %s",
                     configlist[i].start_hour_min, configlist[i].end_hour_min,
                     time_hour_min_sec, configlist[i].cmd, configlist[i].args);
            sprintf(msg, ":h=%s:a=P-OUTSIDE_OF:l=%d:c=%s:x=%s:", hostname,
                    timer, configlist[i].category, configlist[i].cmd);
            sendprobemessage(msg, i);
          }
        }
      }
    }
  }
}

/*
 * show high pcpu processes
 */
void
highpcpu(void)
{ int i;
  char msg[1024];

  for (i = 0; i < lastconfig; i++) {
    if (configlist[i].start_hour_min == configlist[i].end_hour_min)
      continue;
    if ((configlist[i].weekday & weekday) == 0)
      continue;

    if (configlist[i].start_hour_min <= time_hour_min &&
        time_hour_min < configlist[i].end_hour_min) {
      if (ishighpcpu(i)) {
        if (1 <= f.debug) {
          if (configlist[i].alarm_level != 0)
            fprintf(outfile, "high pcpu at %06d: %s %s {%04d %04d %d}\n",
                    time_hour_min_sec, configlist[i].cmd, configlist[i].args,
                    configlist[i].start_hour_min, configlist[i].end_hour_min,
                    configlist[i].alarm_level);
        }
        else {
          /*
           * use time_hour_min_sec to be sure that all messages will be
           * differents. They will not be buffered by a distant syslogd
           */
          if (configlist[i].alarm_level != 0) {
            if (1 <= f.syslog)
              syslog(LOG_ERR, "high pcpu at %06d: %s %s {%04d %04d %d}",
                     time_hour_min_sec, configlist[i].cmd, configlist[i].args,
                     configlist[i].start_hour_min, configlist[i].end_hour_min,
                     configlist[i].alarm_level);
            sprintf(msg, ":h=%s:a=P-PCPU:l=%d:c=%s:x=%s:", hostname,
                    timer, configlist[i].category, configlist[i].cmd);
            sendprobemessage(msg, i);
          }
        }
      }
    }
  }
}

/*
 * show high etime processes
 */
void
highetime(void)
{ int i;
  char msg[1024];

  for (i = 0; i < lastconfig; i++) {
    if (configlist[i].start_hour_min == configlist[i].end_hour_min)
      continue;
    if ((configlist[i].weekday & weekday) == 0)
      continue;

    if (configlist[i].start_hour_min <= time_hour_min &&
        time_hour_min < configlist[i].end_hour_min) {
      if (ishighetime(i)) {
        if (1 <= f.debug) {
          if (configlist[i].alarm_level != 0)
            fprintf(outfile, "high etime at %06d: %s %s {%04d %04d %d}\n",
                    time_hour_min_sec, configlist[i].cmd, configlist[i].args,
                    configlist[i].start_hour_min, configlist[i].end_hour_min,
                    configlist[i].alarm_level);
        }
        else {
          /*
           * use time_hour_min_sec to be sure that all messages will be
           * differents. They will not be buffered by a distant syslogd
           */
          if (configlist[i].alarm_level != 0) {
            if (1 <= f.syslog)
              syslog(LOG_ERR, "high etime at %06d: %s %s {%04d %04d %d}",
                     time_hour_min_sec, configlist[i].cmd, configlist[i].args,
                     configlist[i].start_hour_min, configlist[i].end_hour_min,
                     configlist[i].alarm_level);
            sprintf(msg, ":h=%s:a=P-ETIME:l=%d:c=%s:x=%s:", hostname,
                    timer, configlist[i].category, configlist[i].cmd);
            sendprobemessage(msg, i);
          }
        }
      }
    }
  }
}

/*
 * show high vsz processes
 */
void
highvsz(void)
{ int i;
  char msg[1024];

  for (i = 0; i < lastconfig; i++) {
    if (configlist[i].start_hour_min == configlist[i].end_hour_min)
      continue;
    if ((configlist[i].weekday & weekday) == 0)
      continue;

    if (configlist[i].start_hour_min <= time_hour_min &&
        time_hour_min < configlist[i].end_hour_min) {
      if (ishighvsz(i)) {
        if (1 <= f.debug) {
          if (configlist[i].alarm_level != 0)
            fprintf(outfile, "high vsz at %06d: %s %s {%04d %04d %d}\n",
                    time_hour_min_sec, configlist[i].cmd, configlist[i].args,
                    configlist[i].start_hour_min, configlist[i].end_hour_min,
                    configlist[i].alarm_level);
        }
        else {
          /*
           * use time_hour_min_sec to be sure that all messages will be
           * differents. They will not be buffered by a distant syslogd
           */
          if (configlist[i].alarm_level != 0) {
            if (1 <= f.syslog)
              syslog(LOG_ERR, "high vsz at %06d: %s %s {%04d %04d %d}",
                     time_hour_min_sec, configlist[i].cmd, configlist[i].args,
                     configlist[i].start_hour_min, configlist[i].end_hour_min,
                     configlist[i].alarm_level);
            sprintf(msg, ":h=%s:a=P-VSZ:l=%d:c=%s:x=%s:", hostname,
                    timer, configlist[i].category, configlist[i].cmd);
            sendprobemessage(msg, i);
          }
        }
      }
    }
  }
}

/*
 * remove processes left from a previous loop
 */
void
removeoldprocesses(void)
{ int i;

  i = 0;
  while (i < lastproc) {
    if (proclist[i].loop != loop) {
      if (2 <= f.debug && record == 0)
        fprintf(outfile, "- %04d %04d %.1f %u %u %s %s %s\n",
                proclist[i].start_hour_min, time_hour_min,
                proclist[i].pcpu, proclist[i].etime, proclist[i].vsz,
                proclist[i].pid, proclist[i].cmd, proclist[i].args);
      free(proclist[i].pid);
      free(proclist[i].cmd);
      free(proclist[i].args);
      if (proclist[i].match != -1) {
	configlist[proclist[i].match].match = -1;
        proclist[i].match = -1;
      }
      lastproc--;
      if (i != lastproc) {
        proclist[i].pid = proclist[lastproc].pid;
        proclist[i].pcpu= proclist[lastproc].pcpu;
        proclist[i].etime= proclist[lastproc].etime;
        proclist[i].vsz = proclist[lastproc].vsz;
        proclist[i].cmd = proclist[lastproc].cmd;
        proclist[i].args= proclist[lastproc].args;
        proclist[i].loop= proclist[lastproc].loop;
        proclist[i].start_hour_min = proclist[lastproc].start_hour_min;
        proclist[i].match = proclist[lastproc].match;
        if (proclist[i].match != -1) {
          configlist[proclist[i].match].match = i;
        }
        proclist[i].state = proclist[lastproc].state;
        i--;
      }
    }
    i++;
  }
}

/*
 * compute etime from ps(1) output
 */
unsigned int
etimehourminsec(char *buffer)
{ char *ptr;
  unsigned int number, number2, totalresult;

  /* read Solaris and Linux etime format [[d+-]hh:]mm:ss */
  /* read FreeBSD etime format m+:ss */

  /* printf("call etimehourminsec with: %s\n", buffer);*/

  ptr = buffer;
  number = 0;
  number2 = 0;

  while (*ptr) {
    /*printf ("running with %c\n", *ptr);*/
    if (isdigit((int)*ptr)) {
      number *= 10;
      number += (int)(*ptr - '0');
      /*printf ("found digit = %c\n", *ptr);*/
    }
#if defined(__FreeBSD__)
    if (*ptr == ':') {
      /*printf ("found - with number = %d\n", number);*/
      number2 = number;
      number = 0;
    }
#else
    if (*ptr == '-') {
      /*printf ("found - with number = %d\n", number);*/
      number2 = number;
      number = 0;
    }
#endif
    ptr++;
  }
  /*printf ("found $ with number = %d\n", number);*/

#if defined(__FreeBSD__)
  totalresult = (number2 / 60) * 10000 + (number2 % 60) * 100 + number;
#else
  totalresult = number2 * 240000 + number;
#endif

  if (2 <= f.debug)
    fprintf(outfile, "etime computed %d\n", totalresult);

  return (totalresult);
}

/*
 * compute a line of ps(1) output
 */
void
processline(char *buffer)
{ char *pid, *cmd,  *args;
  char *remainder;
  double pcpu;
  unsigned int etime;
  unsigned int vsz;
  int i;

  chomp(buffer);

  if (4 <= f.debug)
    fprintf (outfile, "processline(): ##%s##\n", buffer);

  remainder = buffer;
  skipwhitespace(&remainder);
  pid = shift(&remainder, 1 /* stop at first space */);
  pcpu = atof(shift(&remainder, 1 /* stop at first space */));
  etime = etimehourminsec(shift(&remainder, 1 /* stop at first space */));
  vsz = atoi(shift(&remainder, 1 /* stop at first space */));
  cmd = shift(&remainder, 1 /* stop at first space */);
  args= shift(&remainder, 0 /* do not stop at space */);

  for (i = 0; i < lastproc; i++) {
    if (strcmp(proclist[i].pid, pid) == 0 &&
        strcmp(proclist[i].cmd, cmd) == 0 &&
        strcmp(proclist[i].args, args) == 0) {
      proclist[i].loop = loop;
      return;
    }
  }
  if (lastproc == MAXLASTPROC)
    myerrx(1, "increase MAXLASTPROC");
  proclist[lastproc].pid = strdup(pid);
  proclist[lastproc].pcpu= pcpu;
  proclist[lastproc].etime= etime;
  proclist[lastproc].vsz= vsz;
  proclist[lastproc].cmd = strdup(cmd);
  proclist[lastproc].args= strdup(args);
  proclist[lastproc].loop= loop;
  proclist[lastproc].start_hour_min = time_hour_min;
  proclist[lastproc].match = -1;
  proclist[lastproc].state = STATEUNKNOWN;
  lastproc++;
  if (2 <= f.debug && record == 0)
    fprintf(outfile, "+ %04d 2400 %.1f %u %u %s %s %s\n",
            time_hour_min, pcpu, etime, vsz, pid, cmd, args);
}

void
process(void)
{
  /* remove no longer running processes (those with a not up to date tag) */
  removeoldprocesses();
  /* unmark processes : set to state STATEUNKNOWN */
  resetstate();
  /* alarm on missing processes */
  missing(); /* set to state STATERUNNING when found */
  /* alarm on running processus during wrong period */
  running(); /* set to state STATEWRONGTIME when found */
  /* alarm on high pcpu processes */
  highpcpu();
  /* alarm on high etime processes */
  highetime();
  /* alarm on high vsz processes */
  highvsz();
  /* alarm on running processus left (not in config file) */
  if (f.excluded) {
    markexcluded();
    unknown();
  }
  if (f.itrs)
    dumpconfigtoitrs();
  if (f.html)
    dumpconfigtohtml();
}

void
usage(void)
{
  fprintf(stderr, "%s\n%s\n",
      "usage: pschecker [-1cdehiqsv] [-E excludefilename] [-o outfile]",
      "                 [-p probehost] [-R time] [-t timer] conffilename");
  exit(1);
}

void
terminate(int notused)
{
  sig_terminate = 1;
}

void
initprobe(void)
{ struct hostent *hp;
  struct servent *sserv;
  int i;

  /* probehost is set by readconffile() or by -p option flag */
  for (i = 0; i < lastprobehost; i++) {
    if (f.debug != 0)
      fprintf(outfile, "probehost set to ##%s## with public=%d\n",
          probehostlist[i].probehost, probehostlist[i].public);
    if ((probehostlist[i].sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
      myerrx(1, "socket");

    hp = gethostbyname(probehostlist[i].probehost);
    if (hp == NULL) {
      fprintf(stderr, "#probehost %s directive found in config file\n", probehostlist[i].probehost);
      myerrx(1, "probehost: host not found in hosts database");
    }

    sserv = getservbyname("phchaprobe", "udp");
    if (sserv == NULL)
      myerrx(1, "phchaprobe/udp: service not found in services database");

    bcopy(hp->h_addr, &(probehostlist[i].name).sin_addr, hp->h_length);
    (probehostlist[i].name).sin_family = AF_INET;
    (probehostlist[i].name).sin_port = sserv->s_port;
  }
}

void
doneprobe(void)
{ int i;

  for (i = 0; i < lastprobehost; i++)
    close(probehostlist[i].sockfd);
}

int
main (int argc, char **argv)
{ char buf[MAXBUFFERSIZE];
  int fildes[2];
  int ch;
  time_t thetime;
  struct tm *ptm;
  char msg[1024];
  pid_t pid_child;
  FILE *stream;
  int status;

  loop = 1;
  lastproc = 0;
  lastprobehost = 0;
  time_hour_min = 0;
  time_hour_min_sec = 0;
  weekday = atobin("1111111");
  category = strdup("n/a");
  f.debug = 0;
  f.html = 0;
  f.itrs = 0;
  f.once = 0;
  f.syslog = 0;
  f.displayconfig = 0;
  f.quiet = 0;
  f.sort_by_start_time = 0;
  f.excluded = 0;
  excludefilename = NULL;
  outfile = stdout;
  sig_terminate = 0;
  timer = 30;
  record = 0;
  processesallowed = 100;

  while ((ch = getopt(argc, argv, "1E:R:cdehio:p:qst:v")) != EOF)
    switch(ch) {
    case '1' : f.once = 1;
               break;
    case 'c' : f.displayconfig++;
               break;
    case 'd' : f.debug++;
               break;
    case 'E' : excludefilename = optarg;
               break;
    case 'e' : f.excluded++;
               break;
    case 'h' : f.itrs = 0; f.html++;
               break;
    case 'i' : f.html = 0; f.itrs++;
               break;
    case 'o' : outfile = fopen(optarg, "w");
               if (outfile == NULL)
                 myerrx(1, "incorrect output file name");
               break;
    case 'p' : if (lastprobehost == MAXPROBELIST)
                 myerrx(1, "increase MAXPROBELIST");
               probehostlist[lastprobehost].public = -1;
               probehostlist[lastprobehost].probehost = optarg;
               lastprobehost++;
               break;
    case 'q' : f.quiet++;
               break;
    case 'R' : record = atoi(optarg);
               if (record < 0 || 2400 < record)
                 myerrx(1, "-R value not in [0000..2400]");
               break;
    case 's' : f.syslog++;
               break;
    case 't' : timer = atoi(optarg);
               break;
    case 'v' : printf("%s\n", rcsid);
               exit(1);
               break;
    default  : usage();
    }
  argc -= optind;
  argv += optind;

  if ((conffilename = argv[0]) == NULL)
    usage();

  signal(SIGINT, terminate);
  signal(SIGTERM, terminate);

  if (f.debug == 0 && 1 <= f.syslog)
    openlog("pschecker", 0 /*LOG_PID*/, LOG_USER);

  virtualsize = getmemorysize() / 2;
  lastexclude = 0;
  excludedprocesscount = 100;
  readconffile();
  if (f.displayconfig != 0) {
    displayconfig();
    if (f.debug == 0 && 1 <= f.syslog)
      closelog();
    return (0);
  }
  excludedprocesscount = 100;
  readexcludefile();

  gethostname(hostname, sizeof(hostname));

  initprobe();

  while (! sig_terminate) {
    pipe(fildes);

    switch((pid_child = fork())) {
       case -1 : myerrx(1, "fork failed");
       case 0  : /* child */
                 close(fildes[0]);      /* close the half pipe we don't use */
                 close(STDOUT_FILENO);  /* and stdout */
                 dup2(fildes[1], STDOUT_FILENO); /* stdout -> pipe */
#if defined(__FreeBSD__)
                 execl("/bin/ps", "ps", "-axo", "pid,pcpu,etime,vsz,command", NULL);
#elif defined(__linux__)
                 if (f.sort_by_start_time == 1)
                   execl("/bin/ps", "ps", "-eo", "pid,pcpu,etime,vsz,command", "--sort", "start_time", NULL);
                 else
                   execl("/bin/ps", "ps", "-eo", "pid,pcpu,etime,vsz,command", NULL);
#else
                 execl("/usr/bin/ps", "ps", "-eo", "pid,pcpu,etime,vsz,args", NULL);
#endif
                 /* NOTREACHED */
                 myerrx(0, "exec failed");
       default : /* parent */
                 close(fildes[1]);      /* close the half pipe we don't use */
                 if ((stream = fdopen(fildes[0], "r")) == NULL) {
                   kill(pid_child, SIGTERM);
                   exit(1);
                 }

                 processes = 0;
                 while (fgets(buf, MAXBUFFERSIZE, stream)) {
                   processline(buf);
                   processes++;
                 }

                 fclose(stream);
                 close(fildes[0]);

                 if (waitpid(pid_child, &status, 0) == -1)
                   myerrx(1, "waitpid() failed");

                 thetime = time(NULL);
                 ptm = localtime(&thetime);
                 time_hour_min = ptm->tm_hour*100 + ptm->tm_min;
                 time_hour_min_sec = time_hour_min*100 + ptm->tm_sec;
                 weekday = 1 << (6 - ptm->tm_wday);
                 if (f.debug == 0) {
                   sprintf(msg, ":h=%s:a=P-PROBE:l=%d:", hostname, timer);
                   sendprobemessage(msg, -2);
                 }
                 if (processesallowed < processes) {
                   if (1 <= f.syslog)
                     syslog(LOG_ERR, "high number of process at %06d: %d/%d ",
                       time_hour_min_sec, processes, processesallowed);
                   sprintf(msg, ":h=%s:a=P-NUMPROCESS:l=%d:x=%d/%d:", hostname, timer, processes, processesallowed);
                   sendprobemessage(msg, -2);
                 }
                 process();
                 dumpall();
                 break;
    }
    if (2 <= f.debug)
      fprintf(outfile, "===hour: %04d==loop: %d=="
              "=( )list (+)add (-)remove===\n", time_hour_min, loop);
    loop++;
    if (f.once || (record != 0 && record <= time_hour_min))
      sig_terminate = 1;
    else
      sleep(timer);
  }

  fclose(outfile);
  if (f.debug == 0 && 1 <= f.syslog)
    closelog();

  doneprobe();
  return (0);
}
